# Standard library imports
import sys
from logging import getLogger
from os import environ
from pathlib import Path
from typing import Annotated

# Third party imports
from aeth_ext.settings import BaseSettings
from pydantic import Field
from pydantic_settings import SettingsConfigDict

logger = getLogger(__name__)

environ.setdefault("PYDANTIC_ERRORS_INCLUDE_URL", "false")


__all__ = ["Settings"]

CWD = Path(__file__).parent if getattr(sys, "frozen", False) else Path.cwd()


class Settings(BaseSettings):
  model_config = (
    SettingsConfigDict(
      env_file=CWD / ".env",
      env_file_encoding="utf-8",
      env_ignore_empty=True,
      extra="ignore",
    )
    if __debug__
    else SettingsConfigDict()
  )

  persisted_dir_loc: Annotated[Path, Field(alias="PERSISTED_DIR_LOC")] = (
    CWD / "persisted_data" if __debug__ else Path("/app/persisted_data")
  )

  timeclock_employee_input_loc: Annotated[Path, Field(alias="TIMECLOCK_EMPLOYEE_INPUT_LOC")] = (
    persisted_dir_loc / "timeclock_employee_input"
  )
  timeclock_font_input_loc: Annotated[Path, Field(alias="TIMECLOCK_FONT_INPUT_LOC")] = persisted_dir_loc / "timeclock_font_input"

  allotted_hours_sheet_id: Annotated[str, Field(alias="ALLOTTED_HOURS_SHEET_ID")] = "1Fn1FBZZAQwrB6v-wkMGkeIN12Aui7SyZvYpEBvc4Wjk"

  @property
  def sft_website_creds_file(self) -> Path:
    return self._creds_file_reusable("SFT website creds file not found at expected location", "secrets", "sft_ftp_creds.json")

  @property
  def sas_ftp_creds_file(self) -> Path:
    return self._creds_file_reusable("SAS FTP creds file not found at expected location", "secrets", "sas_ftp_creds.json")

  @property
  def ryo_ftp_creds_file(self) -> Path:
    return self._creds_file_reusable("RYO FTP creds file not found at expected location", "secrets", "ryo_ftp_creds.json")

  @property
  def google_api_key_file(self) -> Path:
    return self._creds_file_reusable(
      "Google API key file not found at expected location", "secrets", "scheduledreportaggregator-bdd6c704c6b1.json"
    )



# Standard library imports
from datetime import timedelta
from logging import getLogger
from re import compile
from typing import TYPE_CHECKING, override
from zoneinfo import ZoneInfo

# Third party imports
import apscheduler.executors.base as exec_base
from apscheduler.events import EVENT_JOB_ADDED, EVENT_JOB_EXECUTED, EVENT_JOB_MISSED, JobEvent, JobExecutionEvent
from apscheduler.executors.asyncio import AsyncIOExecutor
from apscheduler.jobstores.base import ConflictingIdError
from apscheduler.jobstores.memory import MemoryJobStore
from apscheduler.schedulers.asyncio import AsyncIOScheduler
from apscheduler.schedulers.base import STATE_RUNNING, STATE_STOPPED
from apscheduler.util import iscoroutinefunction_partial

# First party imports
from aeth_ext.errors.err_handling import handle_fatal_exc_sync
from aeth_ext.utils import get_now

# Local folder imports
from .environment_init_vars import SETTINGS

if TYPE_CHECKING:
  # Standard library imports
  from concurrent.futures import Future
  from datetime import datetime
  from typing import Any, TextIO

  # Third party imports
  from apscheduler.job import Job


logger = getLogger(__name__)

__all__ = ["Scheduler"]

DO_NOT_LOG_PATTERNS = [
  compile(r".*?_register_dropoff_.*"),
  compile(r"submit_queued_writes_to_pool"),
  compile(r".*?_register_pickup_.*"),
  compile(r"print_jobs"),
  compile(r"heartbeat"),
]


def run_job(job: Job, jobstore_alias: str, run_times: list[datetime], logger_name: str):
  """
  Called by executors to run the job. Returns a list of scheduler events to be dispatched by the
  scheduler.

  """
  events = []
  local_logger = getLogger(logger_name)
  for run_time in run_times:
    # See if the job missed its run time window, and handle
    # possible misfires accordingly
    if job.misfire_grace_time is not None:  # pyright: ignore[reportUnnecessaryComparison]
      now = get_now(ZoneInfo("UTC"))

      difference = now - run_time
      grace_time = timedelta(seconds=job.misfire_grace_time)
      if difference > grace_time:
        events.append(JobExecutionEvent(EVENT_JOB_MISSED, job.id, jobstore_alias, run_time))
        local_logger.warning(f'Run time of job "{job.id}" was missed by {difference}')
        continue

    if not any(pattern.match(job.id) for pattern in DO_NOT_LOG_PATTERNS):
      logger.info(f'Scheduler: Running job "{job.id}" (scheduled at {run_time})')
      local_logger.info(f'Scheduler: Running job "{job.id}" (scheduled at {run_time})')
    else:
      logger.debug(f'Scheduler: Running job "{job.id}" (scheduled at {run_time})')
      local_logger.debug(f'Scheduler: Running job "{job.id}" (scheduled at {run_time})')
    retval = job.func(*job.args, **job.kwargs)
    events.append(JobExecutionEvent(EVENT_JOB_EXECUTED, job.id, jobstore_alias, run_time, retval=retval))
    if not any(pattern.match(job.id) for pattern in DO_NOT_LOG_PATTERNS):
      local_logger.info(f'Job "{job.id}" executed successfully')
    else:
      local_logger.debug(f'Job "{job.id}" executed successfully')

  return events


async def run_coroutine_job(job: Job, jobstore_alias: str, run_times: list[datetime], logger_name: str):
  """Coroutine version of run_job()."""
  events = []
  local_logger = getLogger(logger_name)
  for run_time in run_times:
    # See if the job missed its run time window, and handle possible misfires accordingly
    if job.misfire_grace_time is not None:  # pyright: ignore[reportUnnecessaryComparison]
      now = get_now(ZoneInfo("UTC"))

      difference = now - run_time
      grace_time = timedelta(seconds=job.misfire_grace_time)
      if difference > grace_time:
        events.append(JobExecutionEvent(EVENT_JOB_MISSED, job.id, jobstore_alias, run_time))
        local_logger.warning(f'Run time of job "{job.id}" was missed by {difference}')
        continue

    if not any(pattern.match(job.id) for pattern in DO_NOT_LOG_PATTERNS):
      logger.info(f'Scheduler: Running job "{job.id}" (scheduled at {run_time})')
      local_logger.info(f'Scheduler: Running job "{job.id}" (scheduled at {run_time})')
    else:
      logger.debug(f'Scheduler: Running job "{job.id}" (scheduled at {run_time})')
      local_logger.debug(f'Scheduler: Running job "{job.id}" (scheduled at {run_time})')
    retval = await job.func(*job.args, **job.kwargs)
    events.append(JobExecutionEvent(EVENT_JOB_EXECUTED, job.id, jobstore_alias, run_time, retval=retval))
    if not any(pattern.match(job.id) for pattern in DO_NOT_LOG_PATTERNS):
      local_logger.info(f'Job "{job.id}" executed successfully')
    else:
      local_logger.debug(f'Job "{job.id}" executed successfully')

  return events


exec_base.run_job = run_job
exec_base.run_coroutine_job = run_coroutine_job


class CustomAsyncIOExecutor(AsyncIOExecutor):
  @override
  def _do_submit_job(self, job: Job, run_times: list[datetime]):
    @handle_fatal_exc_sync
    def callback(f: Future[Any]):
      self._pending_futures.discard(f)
      # try:
      events = f.result()
      # except BaseException as e:
      #   self._run_job_error(job.id, *exc_info()[1:])
      #   raise e
      # else:
      self._run_job_success(job.id, events)

    if iscoroutinefunction_partial(job.func):
      coro = run_coroutine_job(job, job._jobstore_alias, run_times, self._logger.name)
      f = self._eventloop.create_task(coro)
    else:
      f = self._eventloop.run_in_executor(None, run_job, job, job._jobstore_alias, run_times, self._logger.name)

    f.add_done_callback(callback)
    self._pending_futures.add(f)


class Scheduler(AsyncIOScheduler):
  @classmethod
  def init_scheduler(cls) -> Scheduler:

    job_stores = {
      "default": MemoryJobStore(),
      "system_jobs": MemoryJobStore(),
      "general_jobs": MemoryJobStore(),
    }

    job_defaults = {
      "misfire_grace_time": 60,
      "coalesce": True,
    }

    executors = {
      "default": CustomAsyncIOExecutor(),
    }

    return cls(
      executors=executors,
      jobstores=job_stores,
      job_defaults=job_defaults,
      daemon=False,
      timezone=SETTINGS.tz,
    )

  @override
  def _real_add_job(self, job: Job, jobstore_alias: str, replace_existing: bool):
    """
    :param Job job: the job to add
    :param bool replace_existing: ``True`` to use update_job() in case the job already exists
        in the store

    """
    replacements = {key: value for key, value in self._job_defaults.items() if not hasattr(job, key)}
    # Calculate the next run time if there is none defined
    if not hasattr(job, "next_run_time"):
      now = get_now(self.timezone)

      replacements["next_run_time"] = job.trigger.get_next_fire_time(None, now)

    # Apply any replacements
    job._modify(**replacements)

    # Add the job to the given job store
    store = self._lookup_jobstore(jobstore_alias)
    try:
      store.add_job(job)
    except ConflictingIdError:
      if replace_existing:
        store.update_job(job)
      else:
        raise

    # Mark the job as no longer pending
    job._jobstore_alias = jobstore_alias

    # Notify listeners that a new job has been added
    event = JobEvent(EVENT_JOB_ADDED, job.id, jobstore_alias)
    self._dispatch_event(event)

    self._logger.info(f'Added job "{job.id}" to job store')

    # Notify the scheduler about the new job
    if self.state == STATE_RUNNING:
      self.wakeup()

  @override
  def print_jobs(self, jobstore: str | None = None, out: TextIO | None = None):
    """
    print_jobs(jobstore=None, out=sys.stdout)

    Prints out a textual listing of all jobs currently scheduled on either all job stores or
    just a specific one.

    :param str|unicode jobstore: alias of the job store, ``None`` to list jobs from all stores
    :param file out: a file-like object to print to (defaults to  **sys.stdout** if nothing is
        given)

    """
    lines = []
    with self._jobstores_lock:
      if self.state == STATE_STOPPED:
        lines.append("Pending jobs:")
        if self._pending_jobs:
          lines.extend(f"  {job.id}" for job, jobstore_alias, _ in self._pending_jobs if jobstore in (None, jobstore_alias))
        else:
          lines.append("  No pending jobs")
      else:
        for alias, store in sorted(self._jobstores.items()):
          if jobstore in (None, alias):
            lines.append(f"Jobstore {alias}:")
            if jobs := store.get_all_jobs():
              lines.extend(f"  {job.id}" for job in jobs)
            else:
              lines.append("  No scheduled jobs")

    logger.debug("\n".join(lines))



if __name__ == "__main__":
  # First party imports
  from aeth_ext.logging.init import init_logging

  init_logging()

# Standard library imports
from atexit import register
from datetime import date
from decimal import Decimal
from json import load
from logging import getLogger
from pathlib import Path
from subprocess import run
from sys import executable
from typing import TYPE_CHECKING, NamedTuple, override

# Third party imports
from google.oauth2.service_account import Credentials
from gspread.auth import authorize
from gspread.http_client import BackOffHTTPClient
from gspread.utils import DateTimeOption, Dimension, ValueRenderOption
from pandas import notna, read_csv

# First party imports
from aeth_ext.types import EmailMessageParts
from aeth_ext.utils import batch_send_emails, prepare_email_message
from scheduled_report_aggregator.environment_init_vars import CWD, SETTINGS
from scheduled_report_aggregator.jobs.base import CanRescheduleJobError, CronArgs, JobBase

# Local folder imports
from .allotted_hours_model import AllottedHoursModel

if TYPE_CHECKING:
  # Third party imports
  from gspread.client import Client

  # First party imports
  from scheduled_report_aggregator.custom_types import StoreNum

logger = getLogger(__name__)

__all__ = ["TimeclockJob"]

TIMECLOCK_PLAYGROUND = CWD / "timeclock_playground"
SUBPROCESS_EMPLOYEE_INPUT = TIMECLOCK_PLAYGROUND / "employee_input"
SUBPROCESS_FONT_INPUT = TIMECLOCK_PLAYGROUND / "font_input"
SUBPROCESS_OUTPUT_FOLDER = TIMECLOCK_PLAYGROUND / "output"

SUBPROCESS_PERSISTED_DATA_FOLDER = TIMECLOCK_PLAYGROUND / "persisted_data"

SUBPROCESS_MANIFEST_PATH = TIMECLOCK_PLAYGROUND / "manifest.json"


MANUAL_EMPLOYEE_LIST_CSV = max(SETTINGS.timeclock_employee_input_loc.iterdir(), key=lambda f: f.stat().st_mtime)


WEEK_DATA_EXPECTED_COLUMNS = (
  "Store",
  "Employee Name",
  "In Time",
  "Out Time",
  "Time Worked",
  "Hours Worked",
  "Date",
  "Store Number",
)

EMPLOYEE_INFO_EXPECTED_COLUMNS = (
  "Employee #",
  "First Name",
  "Last Name",
  "Status",
  "Manager",
  "Hire Date",
  "Group",
  "Work Type",
  "Notes",
  "Created",
  "Updated",
)


class ManifestEntry(NamedTuple):
  csv: Path
  pdf: Path


class OverUnderEntry(NamedTuple):
  store: StoreNum
  week_ending: date
  alloted_hours: int
  worked_hours: Decimal
  over_under_hours: Decimal  # positive for over, negative for under
  pdf_path: Path
  csv_path: Path


DEFAULT_SCOPES = ["https://www.googleapis.com/auth/spreadsheets"]


class TimeclockJob(JobBase):
  creds = Credentials.from_service_account_file(SETTINGS.google_api_key_file, scopes=DEFAULT_SCOPES)

  sheet_tab_store_range = "'Sheet1'!R2C1:C1"
  sheet_tab_allotted_hours_range = "'Sheet1'!R2C3:C3"

  email_recipients = (
    # "denirosaco@sweetfiretobacco.com",
    # "jacob.ogden@sweetfiretobacco.com",
    # "franksaco@sweetfiretobacco.com",
    # "spt@sweetfiretobacco.com",
    # "pauline@sweetfiretobacco.com",
    "timeclockhoursreports@sweetfiretobacco.com",
  )

  def __post_init__(self) -> None:
    self.output_folder = self.job_holding_folder / "output"
    self.manifest_path = self.job_holding_folder / "manifest.json"

    self.manifest_path.touch(exist_ok=True)
    self.output_folder.mkdir(exist_ok=True)
    TIMECLOCK_PLAYGROUND.mkdir(exist_ok=True)

    # ensure manifest file is empty
    self.manifest_path.write_text("")

    SUBPROCESS_EMPLOYEE_INPUT.unlink(missing_ok=True)
    SUBPROCESS_FONT_INPUT.unlink(missing_ok=True)
    SUBPROCESS_OUTPUT_FOLDER.unlink(missing_ok=True)
    SUBPROCESS_PERSISTED_DATA_FOLDER.unlink(missing_ok=True)
    SUBPROCESS_MANIFEST_PATH.unlink(missing_ok=True)

    SUBPROCESS_EMPLOYEE_INPUT.symlink_to(SETTINGS.timeclock_employee_input_loc, target_is_directory=True)
    SUBPROCESS_FONT_INPUT.symlink_to(SETTINGS.timeclock_font_input_loc, target_is_directory=True)
    SUBPROCESS_OUTPUT_FOLDER.symlink_to(self.output_folder, target_is_directory=True)
    SUBPROCESS_PERSISTED_DATA_FOLDER.symlink_to(SETTINGS.persisted_dir_loc, target_is_directory=True)
    SUBPROCESS_MANIFEST_PATH.symlink_to(self.manifest_path, target_is_directory=False)

    register(SUBPROCESS_EMPLOYEE_INPUT.unlink, missing_ok=True)
    register(SUBPROCESS_FONT_INPUT.unlink, missing_ok=True)
    register(SUBPROCESS_OUTPUT_FOLDER.unlink, missing_ok=True)
    register(SUBPROCESS_PERSISTED_DATA_FOLDER.unlink, missing_ok=True)
    register(SUBPROCESS_MANIFEST_PATH.unlink, missing_ok=True)

    self.schedule_local_holding = self.job_holding_folder / "schedule_reports"
    self.schedule_local_holding.mkdir(exist_ok=True)

    self.reports_pickup_folder = self.reports_pickup_base_folder / "Automated Schedule Report"

  @property
  def client(self) -> Client:
    return authorize(self.creds, http_client=BackOffHTTPClient)

  @override
  async def main_job(self) -> None:
    next_report = await self.get_next_timeclock_report()
    manifest_data = self.run_processor(next_report)

    overunder_data = self.calculate_overunder_hours(manifest_data)

    self.send_results(overunder_data)

  async def get_next_timeclock_report(self) -> Path:
    with self.ftp_handlers["sft"].start_session() as ftp:
      remote_files = ftp.listdir(self.reports_pickup_folder.as_posix())

      try:
        youngest_file = max(remote_files, key=lambda f: f.modified_time)
      except ValueError:
        raise CanRescheduleJobError(
          "Error in locating timeclock schedule report", reason=f"No schedule files in {self.reports_pickup_folder}", count_error=False
        ) from None

      if self.check_if_this_week(youngest_file.modified_time.astimezone(SETTINGS.tz)):
        remote_file = self.reports_pickup_folder / youngest_file.filename
        local_file = self.schedule_local_holding / youngest_file.filename

        with local_file.open("wb") as f:
          ftp.download_file(remote_file.as_posix(), f.write)

        return local_file

      else:
        raise CanRescheduleJobError(
          "No schedule files from this week",
          reason=f"Youngest file {youngest_file.filename} modified at {youngest_file.modified_time} is not from this week",
          count_error=False,
        ) from None

  def run_processor(self, csv_file: Path) -> dict[StoreNum, dict[date, ManifestEntry]]:
    # Run timeclock_entry_processor as a subprocess via its CLI.

    self.manifest_path.touch(exist_ok=True)
    self.manifest_path.write_text("")  # ensure manifest is empty before processing

    if __debug__:
      exec_args = [
        executable,
        "-c",
        ";".join(
          [
            "import runpy, sys",
            "sys.argv[0] = 'timeclock_entry_processor'",
            "runpy.run_module('timeclock_entry_processor', run_name='__main__', alter_sys=True)",
          ]
        ),
        str(csv_file),
        str(self.manifest_path),
        str(self.output_folder),
      ]
    else:
      exec_args = [
        executable,
        "-m",
        "timeclock_entry_processor",
        str(csv_file),
        str(self.manifest_path),
        str(self.output_folder),
      ]

    run(exec_args, check=True, cwd=str(TIMECLOCK_PLAYGROUND))

    return self.load_manifest(self.manifest_path)

  def load_manifest(self, manifest_path: Path) -> dict[StoreNum, dict[date, ManifestEntry]]:
    # ensure that manifest file was created and has content
    if not manifest_path.exists() or manifest_path.stat().st_size == 0:
      raise RuntimeError(f"Expected manifest file at {manifest_path} was not created or is empty after processing")

    manifest_data = {}

    with manifest_path.open("r") as f:
      for store, weeks in load(f).items():
        manifest_data[int(store)] = {
          date.fromisoformat(week): ManifestEntry(csv=Path(entry["csv"]), pdf=Path(entry["pdf"])) for week, entry in weeks.items()
        }

    return manifest_data

  def get_allotted_hours(self) -> dict[StoreNum, int]:
    result = self.client.http_client.values_batch_get(
      id=SETTINGS.allotted_hours_sheet_id,
      ranges=[self.sheet_tab_store_range, self.sheet_tab_allotted_hours_range],
      params={
        "majorDimension": Dimension.rows,
        "valueRenderOption": ValueRenderOption.unformatted,
        "dateTimeRenderOption": DateTimeOption.formatted_string,
      },
    )

    allotted_hours = {}

    for store, allotted_hrs in zip(result["valueRanges"][0]["values"], result["valueRanges"][1]["values"], strict=False):
      if not store or not allotted_hrs:
        continue  # skip rows with missing values
      validated_row = AllottedHoursModel.model_validate({"store": store[0], "allotted_hours": allotted_hrs[0]})
      allotted_hours[validated_row.store] = validated_row.allotted_hours
    return allotted_hours

  def calculate_overunder_hours(self, weeks_data: dict[StoreNum, dict[date, ManifestEntry]]) -> set[OverUnderEntry]:
    overunder_data: set[OverUnderEntry] = set()

    allotted_hours = self.get_allotted_hours()

    employee_groups = read_csv(
      MANUAL_EMPLOYEE_LIST_CSV,
      header=0,
      names=EMPLOYEE_INFO_EXPECTED_COLUMNS,
      usecols=(
        "Employee #",
        "Group",
      ),
      index_col="Employee #",
      dtype=str,
    )["Group"].to_dict()

    for store, weeks in weeks_data.items():
      store_allotted_hours = allotted_hours.get(store, 0)

      for week, (week_csv, week_pdf) in weeks.items():
        week_df = read_csv(
          week_csv,
          header=0,
          names=WEEK_DATA_EXPECTED_COLUMNS,
          usecols=(
            # "Store Number",
            "Employee Name",
            "In Time",
            "Out Time",
            "Hours Worked",
          ),
          dtype=str,
        )

        week_df.loc[:, ["employee_id", "first_name", "last_name"]] = week_df["Employee Name"].str.extract(
          r"(?P<employee_id>\d+) - (?P<first_name>\w+) (?P<last_name>\w+)"
        )
        week_df.loc[:, "group"] = week_df["employee_id"].map(employee_groups)

        # map week_df "Hours Worked" to Decimal, treating non-convertible values as 0
        week_df["Hours Worked"] = week_df["Hours Worked"].map(lambda x: Decimal(x) if notna(x) else Decimal(0))

        week_worked_hours = week_df["Hours Worked"].sum()

        overunder_data.add(
          OverUnderEntry(
            store=store,
            week_ending=week,
            alloted_hours=store_allotted_hours,
            worked_hours=week_worked_hours,
            over_under_hours=week_worked_hours - store_allotted_hours,
            pdf_path=week_pdf,
            csv_path=week_csv,
          )
        )

    return overunder_data

  MAX_ALLOWED_OVER_HOURS = Decimal("1")
  MAX_ALLOWED_UNDER_HOURS = Decimal("-10")

  def send_results(self, overunder_data: set[OverUnderEntry]):
    emails_to_send = []

    for store, week_ending, allotted_hours, worked_hours, over_under_hours, pdf_path, csv_path in sorted(
      overunder_data,
      key=lambda x: x.over_under_hours,
    ):
      if self.MAX_ALLOWED_OVER_HOURS >= over_under_hours >= self.MAX_ALLOWED_UNDER_HOURS:
        continue

      over_under_str = "Over" if over_under_hours > 0 else "Under"
      message = (
        f"Store {store} is {over_under_str} allotted hours for the week ending {week_ending.isoformat()}.\n"
        f"  Allotted Hours: {allotted_hours}\n"
        f"  Worked Hours: {worked_hours}\n"
        f"  {over_under_str} Hours: {abs(over_under_hours)}\n"
      )

      emails_to_send.append(
        prepare_email_message(
          EmailMessageParts(
            subject=f"SFT{store:0>3} - Store {over_under_str} Allotted Hours by {abs(over_under_hours): >5} for Week Ending {week_ending.isoformat()}",
            body=message,
            from_addr=SETTINGS.alerts_email,
            to_addrs=self.email_recipients,
            attachments=[pdf_path, csv_path],
          )
        )
      )

    if emails_to_send:
      batch_send_emails(emails_to_send)

  @staticmethod
  def _debug_get_fixed_timeclock_src() -> Path:
    # Standard library imports
    from site import getsitepackages

    # Resolve editable install source without relying on importing the package first.
    for site_dir in getsitepackages():
      site_path = Path(site_dir)
      for pth_file in site_path.glob("__editable__.timeclock_entry_processor-*.pth"):
        src_path = Path(pth_file.read_text(encoding="utf-8").strip())
        if src_path.exists():
          return src_path

    # Fallback: resolve from the installed module if .pth lookup is unavailable.
    # First party imports
    import timeclock_entry_processor

    return Path(timeclock_entry_processor.__file__).resolve().parents[1]

  @staticmethod
  def _create_fixed_env() -> dict[str, str]:
    # Standard library imports
    import os

    child_env = os.environ.copy()
    child_env["PWD"] = str(TIMECLOCK_PLAYGROUND)

    if __debug__:
      timeclock_src = TimeclockJob._debug_get_fixed_timeclock_src()
      existing_pythonpath = child_env.get("PYTHONPATH", "")
      child_env["PYTHONPATH"] = str(timeclock_src) if not existing_pythonpath else f"{timeclock_src}{os.pathsep}{existing_pythonpath}"

    return child_env


if __name__ == "__main__":
  # Third party imports
  import winloop as asyncio

  # First party imports
  # csv_file = CWD / "Time-Clock-Entry-Report_2026-05-14_19-31-12.csv"
  # TimeclockJob().run_processor(csv_file)
  from scheduled_report_aggregator.custom_types import DayOfWeek

  job = TimeclockJob()
  job.init_job(
    scheduler="test",  # pyright: ignore[reportArgumentType]
    job_id="test",
    **CronArgs(day_of_week=DayOfWeek.TUESDAY, hour=9, minute=0, second=0),
  )

  test = job.error_reschedule()

  # result = job.calculate_overunder_hours(job.load_manifest(CWD / "manifest.json"))
  # job.send_results(result)

  asyncio.run(job.main_job())
  pass



# Standard library imports
import os
from logging import getLogger
from pathlib import Path

# Local folder imports
from .environment_settings import Settings

logger = getLogger(__name__)

if os.name != "nt" and hasattr(os, "geteuid") and os.geteuid() == 0:
  logger.warning("Process is running as root on a Unix system. This is not recommended for production.")  # pyright: ignore[reportUnreachable]


__all__ = ["CWD", "SETTINGS"]

# Settings
SETTINGS = Settings.get_settings()

# Folder paths
CWD = Path.cwd()



# Standard library imports
from json import loads
from logging import getLogger
from socket import gaierror
from typing import override

# Third party imports
from paramiko import AutoAddPolicy, SFTPClient, SSHClient

# First party imports
from aeth_ext.ftp import ProtocolEnum, ServerNotAvailableError, SFTPProtocol

# Local folder imports
from .environment_init_vars import SETTINGS

logger = getLogger(__name__)


__all__ = ["RYOSFTPClient", "SASSFTPClient", "SFTSFTPClient"]


class SFTSFTPClient(SFTPProtocol):
  policy = AutoAddPolicy()
  creds = loads(SETTINGS.sft_website_creds_file.read_text())
  KIND = ProtocolEnum.SFTP

  @override
  def get_conn_handler(self) -> SFTPClient:
    try:
      self.ssh_client = SSHClient()
      self.ssh_client.set_missing_host_key_policy(self.policy)

      self.ssh_client.connect(
        hostname=self.creds["HOSTNAME"],
        port=self.creds.get("PORT", 22),
        username=self.creds["USER"],
        password=self.creds["PWD"],
      )
      self.handler = self.ssh_client.open_sftp()
    except ConnectionRefusedError as e:
      raise ServerNotAvailableError(
        f"Could not connect to SFTP server at {self.creds['HOSTNAME']}:{self.creds.get('PORT', 22)}"
        f"\n Server exists but is not running an SFTP service or is blocking the connection."
      ) from e
    except TimeoutError as e:
      raise ServerNotAvailableError(
        f"Connection to SFTP server at {self.creds['HOSTNAME']}:{self.creds.get('PORT', 22)} timed out."
        f"\n Server may be offline or experiencing connectivity issues."
      ) from e
    except gaierror as e:
      raise ServerNotAvailableError(
        f"SFTP server hostname {self.creds['HOSTNAME']} could not be resolved.\n DNS has likely failed"
      ) from e

    return self.handler

  @override
  def close_conn_handler(self) -> None:
    self.handler.close()
    self.ssh_client.close()


class SASSFTPClient(SFTPProtocol):
  policy = AutoAddPolicy()
  creds = loads(SETTINGS.sas_ftp_creds_file.read_text())
  KIND = ProtocolEnum.SFTP

  @override
  def get_conn_handler(self) -> SFTPClient:
    try:
      self.ssh_client = SSHClient()
      self.ssh_client.set_missing_host_key_policy(self.policy)

      self.ssh_client.connect(
        hostname=self.creds["HOSTNAME"],
        port=self.creds.get("PORT", 22),
        username=self.creds["USER"],
        password=self.creds["PWD"],
      )
      self.handler = self.ssh_client.open_sftp()
    except ConnectionRefusedError as e:
      raise ServerNotAvailableError(
        f"Could not connect to SFTP server at {self.creds['HOSTNAME']}:{self.creds.get('PORT', 22)}"
        f"\n Server exists but is not running an SFTP service or is blocking the connection."
      ) from e
    except TimeoutError as e:
      raise ServerNotAvailableError(
        f"Connection to SFTP server at {self.creds['HOSTNAME']}:{self.creds.get('PORT', 22)} timed out."
        f"\n Server may be offline or experiencing connectivity issues."
      ) from e
    except gaierror as e:
      raise ServerNotAvailableError(
        f"SFTP server hostname {self.creds['HOSTNAME']} could not be resolved.\n DNS has likely failed"
      ) from e

    return self.handler

  @override
  def close_conn_handler(self) -> None:
    self.handler.close()
    self.ssh_client.close()


class RYOSFTPClient(SFTPProtocol):
  policy = AutoAddPolicy()
  creds = loads(SETTINGS.ryo_ftp_creds_file.read_text())
  KIND = ProtocolEnum.SFTP

  @override
  def get_conn_handler(self) -> SFTPClient:
    try:
      self.ssh_client = SSHClient()
      self.ssh_client.set_missing_host_key_policy(self.policy)

      self.ssh_client.connect(
        hostname=self.creds["HOSTNAME"],
        port=self.creds.get("PORT", 22),
        username=self.creds["USER"],
        password=self.creds["PWD"],
      )
      self.handler = self.ssh_client.open_sftp()
    except ConnectionRefusedError as e:
      raise ServerNotAvailableError(
        f"Could not connect to SFTP server at {self.creds['HOSTNAME']}:{self.creds.get('PORT', 22)}"
        f"\n Server exists but is not running an SFTP service or is blocking the connection."
      ) from e
    except TimeoutError as e:
      raise ServerNotAvailableError(
        f"Connection to SFTP server at {self.creds['HOSTNAME']}:{self.creds.get('PORT', 22)} timed out."
        f"\n Server may be offline or experiencing connectivity issues."
      ) from e
    except gaierror as e:
      raise ServerNotAvailableError(
        f"SFTP server hostname {self.creds['HOSTNAME']} could not be resolved.\n DNS has likely failed"
      ) from e
    return self.handler

  @override
  def close_conn_handler(self) -> None:
    self.handler.close()
    self.ssh_client.close()



if __name__ == "__main__":
  # Standard library imports
  from sys import platform

  # Third party imports
  from rich.console import Console

  # First party imports
  from aeth_ext import initialize

  RICH_CONSOLE = Console(
    width=None if platform == "win32" else 165,
    log_time=platform == "win32",
  )
  PROJECT_NAME = "ScheduledReportAggregator"
  LOGGING_TYPE = "daily"

  initialize(asyncio=True)
else:
  # Third party imports
  from rich import get_console

  RICH_CONSOLE = get_console()


# Standard library imports
import sys
from asyncio import run, sleep
from datetime import datetime
from logging import getLogger
from typing import TYPE_CHECKING

# Third party imports
from apscheduler.triggers.cron import CronTrigger

# First party imports
from aeth_ext.errors.err_handling import FATAL_EVENT
from scheduled_report_aggregator.custom_types import DayOfWeek
from scheduled_report_aggregator.environment_init_vars import SETTINGS
from scheduled_report_aggregator.jobs import HOLDING_FOLDER, BalanceSheetJob, TimeclockJob
from scheduled_report_aggregator.jobs.base import CronArgs
from scheduled_report_aggregator.scheduler_config import Scheduler

if TYPE_CHECKING:
  # Standard library imports
  from typing import NoReturn

  # First party imports
  from scheduled_report_aggregator.jobs.base import JobBase


logger = getLogger(__name__)


if not __debug__:
  # Heartbeat file for health checks
  HEARTBEAT_FILE = SETTINGS.log_loc_folder / "heartbeat.txt"

  def write_heartbeat():
    """Write current timestamp to heartbeat file for health monitoring."""
    try:
      HEARTBEAT_FILE.write_text(datetime.now(SETTINGS.tz).isoformat())
    except Exception as e:
      logger.error(f"Failed to write heartbeat: {e}")
else:

  def write_heartbeat():
    pass


scheduler = Scheduler.init_scheduler()


jobs: tuple[tuple[type[JobBase], CronArgs], ...] = (
  (TimeclockJob, CronArgs(day_of_week=DayOfWeek.TUESDAY, hour=9, minute=0, second=0)),
  (BalanceSheetJob, CronArgs(day_of_week=DayOfWeek.WEDNESDAY, hour=7, minute=0, second=0)),
)


async def reschedule_jobs() -> None:
  if scheduler.running:
    scheduler.pause()

  scheduler.remove_all_jobs("general_jobs")

  for job_cls, cron_args in jobs:
    job = job_cls.init_job(
      scheduler=scheduler,
      job_id=job_cls.__name__,
      **cron_args,
    )
    job.schedule_registered_jobs()

  if scheduler.running:
    scheduler.resume()


async def main() -> NoReturn:  # sourcery skip: remove-empty-nested-block
  HOLDING_FOLDER.mkdir(exist_ok=True)
  RICH_CONSOLE.rule("[bold red]Booting...[/]", style="bold red")
  # scheduler.add_job(
  #   scheduler.print_jobs,
  #   CronTrigger(minute="*/1"),
  #   id="print_jobs",
  #   replace_existing=True,
  #   jobstore="system_jobs",
  # )

  await reschedule_jobs()  # Schedule all jobs on startup

  # Heartbeat job - writes timestamp every minute for health monitoring
  scheduler.add_job(
    write_heartbeat,
    CronTrigger(minute="*/1"),
    id="heartbeat",
    replace_existing=True,
    jobstore="system_jobs",
  )

  scheduler.add_job(
    reschedule_jobs,
    CronTrigger(
      day_of_week="sun",
      hour=0,
      minute=0,
      second=0,
    ),
    id="reschedule_jobs",
    replace_existing=True,
    jobstore="system_jobs",
  )

  scheduler.start()

  # Write initial heartbeat on startup
  write_heartbeat()

  scheduler.print_jobs()

  # job = TimeclockJob.init_job(
  #   scheduler=scheduler,
  #   job_id=TimeclockJob.__name__,
  #   **CronArgs(day_of_week=DayOfWeek.TUESDAY, hour=14, minute=15, second=0, timezone=SETTINGS.tz),
  # )
  # job.schedule_registered_jobs()

  if __debug__:
    for job_cls, _ in jobs:
      await job_cls().main_job()  # Run each job once immediately in debug mode for testing
    pass

  RICH_CONSOLE.rule("[bold red]Boot Done[/]", style="bold red")
  # with RICH_CONSOLE.status("Application is running."):
  await FATAL_EVENT

  if any(job.errored for job, _ in jobs):
    await sleep(
      600
    )  # Sleep for 10 minutes to allow pending operations from non-error-state processors to flush through before exiting

  try:
    logger.warning("Fatal shutdown: stopping scheduler to freeze application state")
    scheduler.pause()
    scheduler.shutdown(wait=False)
  except Exception as e:
    logger.error(f"Fatal shutdown: failed to stop scheduler cleanly: {e}", exc_info=True)

  sys.exit(1)


if __name__ == "__main__":
  run(main())
